//
// Frequency & Phase Ditector for DigitalPhaseLockedLoop
//
// Copyright(c) 2002 Tatsuyuki Satoh , All rights reserved
//
// Important !
//
// 1. License
//
// This program is freeware for non-commercial use. 
// The user of this source must assume all legal 
// responsibility in the use of this program. 
// This is not GPL.
//
// 2. Guarantee 
//
// An author does no guarantee about this program.
// You can use this under your own risk. 
// 
// 3. A rule to distribute.
//   Do either 1) or 2).
//   1)Include a perfect original source.
//   2)Include all these headers , and add document of 
//     modification points
//
// Histry
// 2002.11.18 Project Start
// 2002.11.22 1st releaes
//

module pll(
//input
  clk_a,div_a,clk_b,div_b,
//output
  clkd_a,clkd_b,
  freq_a,freq_b,
  phase_a,phase_b,
  lock
);

/* configration */
parameter CNT_A_BITS = 14; // CLK_A divider bits
parameter CNT_B_BITS = 14; // CLK_B divider bits

/* module port */
input  clk_a;                   // A Clock Input
input  clk_b;                   // B Clock Input
input  [CNT_A_BITS-1:0] div_a;  // A divide counter (unit = cycle-1)
input  [CNT_B_BITS-1:0] div_b;  // B divide counter (unit = cycle-1)
output clkd_a;                  // A Divided Clock Output
output clkd_b;                  // B Divided Clock Output
output freq_a;                  // A Frequency Ditect (A>>B,for fast loopback)
output freq_b;                  // B Frequency Ditect (A<<B,for fast loopback)
output phase_a;                 // A Phase Ditect (A>B,for slow loopback) *1
output phase_b;                 // B Phase Ditect (A<B,for slow loopback) *2
output lock;                    // Phase Lock (if freqency locked) *3

/*

*1: phase_a doesn't sense when freq_b==H
*2: phase_b doesn't sense when freq_a==H
*3: lock can't use tristate-controll for loopback output pin
    Because it is delayed from phase_a or phase_b signal.

*/

/////////////////////////////////////////////////////////////////////////////
// A/B Divider
/////////////////////////////////////////////////////////////////////////////

reg [CNT_A_BITS-1:0] cnt_a;
reg [CNT_B_BITS-1:0] cnt_b;
reg carry_a , carry_b;

always @(posedge clk_a)
begin
  if(cnt_a==0)
  begin
    cnt_a   <= div_a;
    carry_a <= ~carry_a;
  end else
    cnt_a   <= cnt_a - 1;
end

always @(posedge clk_b)
begin
  if(cnt_b==0)
  begin
    cnt_b   <= div_b;
    carry_b <= ~carry_b;
  end else
    cnt_b   <= cnt_b - 1;
end

/////////////////////////////////////////////////////////////////////////////
// Frequency Compare
/////////////////////////////////////////////////////////////////////////////
reg flock_a; // clk_a:clk_b == 64>61 (under 3%)
reg flock_b; // clk_a:clk_b == 61<64 (over  3%)

parameter FC_BITS = 6;
parameter FC_MAX  = (1<<FC_BITS)-1;
parameter FC_LIMIT= (1<<FC_BITS)-3;

reg ca1,ca2,cb1,cb2;
reg [FC_BITS-1:0] ca,cb;
always @(posedge clk_a)
begin
  ca2 <= ca1;
  ca1 <= carry_a;
  if(~ca2&ca1)
  begin
    if(ca==FC_MAX)
    begin
      flock_a <= 0;
      flock_b <= (cb<FC_LIMIT);
      cb <= 0;
    end
    ca <= ca + 1;
  end else begin
    cb2 <= cb1;
    cb1 <= carry_b;
    if(~cb2&cb1)
    begin
      if(cb==FC_MAX)
      begin
        flock_a <= (ca<FC_LIMIT);
        flock_b <= 0;
        ca <= 0;
      end
      cb <= cb + 1;
    end
  end
end

/////////////////////////////////////////////////////////////////////////////
// Phase Ditector (with freq. rate lock)
/////////////////////////////////////////////////////////////////////////////
reg  ph_a , ph_b;

// Phase A
always @(posedge carry_a or posedge carry_b)
begin
  if(carry_b)
    ph_a <= flock_b;
  else
    ph_a <= ~flock_a;
end

// Phase B
always @(posedge carry_b or posedge carry_a)
begin
  if(carry_a)
    ph_b <= flock_a;
  else
    ph_b <= ~flock_b;
end

/////////////////////////////////////////////////////////////////////////////
// OUTPUT I/F
/////////////////////////////////////////////////////////////////////////////
assign clkd_a  = carry_a;
assign clkd_b  = carry_b;
assign freq_a  = flock_b;
assign freq_b  = flock_a;
assign phase_a = ph_a;
assign phase_b = ph_b;
assign lock    = ~ph_a & ~ph_b & ~freq_a & ~freq_b;

endmodule

